﻿ /*
| Voice Commander by AlienAxeMan (2/21/2023)
| https://www.patreon.com/AlienAxeMan
 \
  | This is a script intended for use within Virt-A-Mate, very handy as a session plugin.
  | Upon scene load, or manual activation via speaking the phrase "Refresh Keywords"," it
  | will scrape the scene Atoms for UI Buttons, and register their text as keywords for the
  | KeywordRecognizer, Invoking the button's action(s) upon recognition of their spoken
  | keyword.  This voice recognition system relies exclusively on the PRIMARY microphone!
  | 
  | 2.0 Notes
  | Added periodic keyword refreshing.
  |
  | 3.0 Notes
  | Optimized keyword refreshing to avoid rebuilding and restarting recognizer unless
  | we detect changes to button keywords.
  |
  | Licensed under CC BY-SA (see https://creativecommons.org/licenses/by-sa/4.0/)
*/

using System;
using UnityEngine;
using System.Collections.Generic;
using System.Text;
using UnityEngine.Windows.Speech;


namespace AlienAxeMan
{
    internal class VoiceCommander : MVRScript
    {
        private string PLUGIN_LABEL = "AlienAxeMan Voice Commander";
        private string PLUGIN_VERSION = "3.0";

        private KeywordRecognizer m_recognizer_internal = null;
        private List<string> m_keywords_internal = null;
        private Dictionary<string, Action> m_keyword_to_action_map_internal;

        private List<string> m_keywords = null;
        private KeywordRecognizer m_recognizer = null;
        private Dictionary<string, UnityEngine.UI.Button> m_keyword_to_button_map;

        private bool m_queued_keyword_refresh = false;

        private JSONStorableBool m_log_keyword_refreshing;
        private JSONStorableBool m_log_keyword_recognition;
        private bool m_auto_refresh_keywords_after_button_invoke = true;

        private JSONStorableBool m_refresh_keywords_periodically;
        private JSONStorableFloat m_refresh_keywords_periodically_rate;
        private float m_last_keyword_refresh_time = 0;


        private void OnSceneLoaded()
        {
            m_queued_keyword_refresh = true;
        }

        private void OnPhraseRecognizedInternal(PhraseRecognizedEventArgs args)
        {
            if (m_log_keyword_recognition.val)
            {
                StringBuilder builder = new StringBuilder();
                builder.AppendFormat("Speech Recognized: {0} ({1})", args.text, args.confidence);
                SuperController.LogMessage(builder.ToString());
            }

            Action action = null;
            if (m_keyword_to_action_map_internal.TryGetValue(args.text, out action))
            {
                action.Invoke();
            }
            else
            {
                SuperController.LogError("Failed to get action for " + args.text);
                foreach (KeyValuePair<string, Action> entry in m_keyword_to_action_map_internal)
                {
                    SuperController.LogError(entry.Key + ": " + entry.Value.ToString());
                }
            }
        }

        private void OnPhraseRecognized(PhraseRecognizedEventArgs args)
        {
            if (m_log_keyword_recognition.val)
            {
                StringBuilder builder = new StringBuilder();
                builder.AppendFormat("Speech Recognized: {0} ({1})", args.text, args.confidence);
                SuperController.LogMessage(builder.ToString());
            }

            UnityEngine.UI.Button button;
            if (m_keyword_to_button_map.TryGetValue(args.text, out button))
            {
                button.onClick.Invoke();

                if (m_auto_refresh_keywords_after_button_invoke)
                {
                    m_queued_keyword_refresh = true;
                }
            } else
            {
                SuperController.LogError("Failed to get button for " + args.text);
            }
        }

        private string SanitizeString(string text)
        {
            // build a copy of the text, starting by replacing any newline characters with a space
            StringBuilder builder = new StringBuilder();
            builder.Append(text.Replace('\n', ' '));

            // strip off any trailing whitespace (including newline characters)
            for (int index = builder.Length - 1; index >= 0; index--)
            {
                if (char.IsWhiteSpace(builder[index]))
                {
                    builder.Remove(index, 1);
                }
                else
                {
                    break;
                }
            }

            return builder.ToString();
        }

        private void AddKeyword(string keyword, UnityEngine.UI.Button button)
        {
            if (m_recognizer != null)
            {
                m_recognizer.Stop();
                m_recognizer.Dispose();
            }

            if (m_keywords.Contains(keyword))
            { 
                // keyword alread exists, skipping due to collision (can't reasonably respond to a keyword affecting more than one button or determine which one the user really wants)
                return;
            }

            m_keywords.Add(keyword);
            m_keyword_to_button_map.Add(keyword, button);

            UnityEngine.UI.Button button_back;
            if (! m_keyword_to_button_map.TryGetValue(keyword, out button_back))
            {
                SuperController.LogError("Failed to get button for " + keyword);
            }

            if (m_recognizer != null)
            {
                m_recognizer = new KeywordRecognizer(m_keywords.ToArray());
                m_recognizer.Start();
            }
        }

        private void RefreshKeywords()
        {
            if (m_log_keyword_refreshing.val)
            {
                SuperController.LogMessage("Refreshing Keywords...");
            }

            Dictionary<string, UnityEngine.UI.Button> detected_keyword_to_button_map = new Dictionary<string, UnityEngine.UI.Button>();

            List<Atom> atoms = GetSceneAtoms();

            // locate all buttons and add them with their label as their keyword
            foreach (var atom in atoms)
            {
                UnityEngine.UI.Button[] buttons = atom.GetComponentsInChildren<UnityEngine.UI.Button>();
                foreach (UnityEngine.UI.Button button in buttons)
                {
                    if (button.name == "Button")
                    {
                        UnityEngine.UI.Text text = button.GetComponentInChildren<UnityEngine.UI.Text>();
                        if (text != null)
                        {
                            string button_text = SanitizeString(text.text);
                            if (button_text != null)
                            {
                                // if we've already found this key, skip any collisions
                                if (detected_keyword_to_button_map.ContainsKey(button_text))
                                {
                                    continue;
                                }

                                detected_keyword_to_button_map.Add(button_text, button);
                            }
                        }
                    }
                }
            }

            // compare the kewords and buttons we just detected with what we already have cached
            bool detected_changes = false;
            foreach (KeyValuePair<string, UnityEngine.UI.Button> entry in detected_keyword_to_button_map)
            {
                // if we don't have this key mapped, or its value is different
                UnityEngine.UI.Button cached_button;
                if (!m_keyword_to_button_map.TryGetValue(entry.Key, out cached_button) || cached_button != entry.Value)
                {
                    detected_changes = true;
                    break;
                }
            }

            // if there are any changes, we need to stop, set the keywords data, and restart the recognizer
            if (detected_changes)
            {
                if (m_recognizer != null)
                {
                    m_recognizer.Stop();
                    m_recognizer.Dispose();
                    m_recognizer = null;
                }

                m_keywords.Clear();
                m_keyword_to_button_map.Clear();

                // if there is at least one keyword (as it's invalid to start a recognizer with no keywords)
                if (detected_keyword_to_button_map.Count > 0)
                {
                    // Note: this is bypassing AddKeyword, and simply copying the entire list and dictionary all at once
                    m_keyword_to_button_map = detected_keyword_to_button_map;
                    m_keywords = m_keyword_to_button_map.Keys.ToList();

                    if (m_log_keyword_refreshing.val)
                    {
                        foreach (var Key in m_keywords)
                        {
                            SuperController.LogMessage("\t" + Key);
                        }
                    }

                    m_recognizer = new KeywordRecognizer(m_keywords.ToArray());
                    m_recognizer.OnPhraseRecognized += OnPhraseRecognized;
                    m_recognizer.Start();
                }
            }
            else // there are no changes, so we can avoid interrupting the keyword recognizer, which may be in the middle of listening to a keyword
            {
                if (m_log_keyword_refreshing.val)
                {
                    SuperController.LogMessage("\tno changes detected");
                }
            }

            m_last_keyword_refresh_time = Time.time;
        }

        private void StateKeywords()
        {
            // TODO: use text to speech and read off the current lists of keywords (both internal and public),
            // adding this action as an internal keyword action, maybe with additional keywords like "Help".
        }

        public void OnEnable()
        {
            if (m_recognizer != null)
            {
                m_recognizer.Start();
            }

            if (m_recognizer_internal != null)
            {
                m_recognizer_internal.Start();
            }
        }

        public void OnDisable()
        {
            if (m_recognizer != null)
            {
                m_recognizer.Stop();
                m_recognizer.Dispose();
                m_recognizer = null;
            }

            if (m_recognizer_internal != null)
            {
                m_recognizer_internal.Stop();
                m_recognizer_internal.Dispose();
                m_recognizer_internal = null;
            }
        }

        public override void Init()
        {
            pluginLabelJSON.val = PLUGIN_LABEL + " " + PLUGIN_VERSION;

            m_keywords_internal = new List<string>();
            m_keywords = new List<string>();
            m_keyword_to_button_map = new Dictionary<string, UnityEngine.UI.Button>();
            m_keyword_to_action_map_internal = new Dictionary<string, Action>();

            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;

            m_log_keyword_refreshing = new JSONStorableBool("Log Keyword Refreshing", false);
            RegisterBool(m_log_keyword_refreshing); // Registering the value is important, so that it will be saved
            CreateToggle(m_log_keyword_refreshing); // the second argument is for 'rightSide'. Supplying true will render the slider on the right-side

            m_log_keyword_recognition = new JSONStorableBool("Log Keyword Recognition", true);
            RegisterBool(m_log_keyword_recognition); // Registering the value is important, so that it will be saved
            CreateToggle(m_log_keyword_recognition); // the second argument is for 'rightSide'. Supplying true will render the slider on the right-side

            m_refresh_keywords_periodically = new JSONStorableBool("Refresh Keywords Periodically", true);
            RegisterBool(m_refresh_keywords_periodically);
            CreateToggle(m_refresh_keywords_periodically, true);

            m_refresh_keywords_periodically_rate = new JSONStorableFloat("Periodic Refresh Rate (in seconds)", 2.0f, 1.0f, 10.0f);
            RegisterFloat(m_refresh_keywords_periodically_rate); // Registering the float is important -- if not they wont save
            UIDynamicSlider slider = CreateSlider(m_refresh_keywords_periodically_rate, true); // the second argument is for 'rightSide'. Supplying true will render the slider on the right-side
            slider.Configure(
                m_refresh_keywords_periodically_rate.name,
                m_refresh_keywords_periodically_rate.min,
                m_refresh_keywords_periodically_rate.max,
                m_refresh_keywords_periodically_rate.val);

            CreateSpacer(true);

            // create the button for refreshing keywords immediately
            UIDynamicButton refresh_keywords_now_button = CreateButton("Refresh Keywords Now", true);
            refresh_keywords_now_button.height = 100;
            refresh_keywords_now_button.button.onClick.AddListener(delegate ()
            {
                m_queued_keyword_refresh = true;
            });

            // expose an action for external callers to be able to manually refresh keywords on demand
            JSONStorableAction m_refresh_keywords_now = new JSONStorableAction("Refresh Keywords Now", RefreshKeywords);
            RegisterAction(m_refresh_keywords_now);
        }
        private void OnDestroy()
        {
            SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
        }

        // Runs once when plugin loads - after Init()
        protected void Start()
        {
            // add internal keywords that aren't dependent on button data (for now it's just a command to refresh the other keywords)
            string refresh_keywords_text = "Refresh Keywords";
            m_keywords_internal.Add(refresh_keywords_text);
            m_keyword_to_action_map_internal.Add(refresh_keywords_text, RefreshKeywords);
            if (m_log_keyword_refreshing.val)
            {
                SuperController.LogMessage("Detected Internal Keyword: " + refresh_keywords_text);
            }

            m_recognizer_internal = new KeywordRecognizer(m_keywords_internal.ToArray());
            m_recognizer_internal.OnPhraseRecognized += OnPhraseRecognizedInternal;
            m_recognizer_internal.Start();

            m_queued_keyword_refresh = true;
        }

        // Regularly polled to check if we have anything in our queue that we should handle now
        private void Update()
        {
            if (m_refresh_keywords_periodically.val)
            {
                float time_since_last_keyword_refresh = Time.time - m_last_keyword_refresh_time;

                if (time_since_last_keyword_refresh >= m_refresh_keywords_periodically_rate.val)
                {
                    m_queued_keyword_refresh = true;
                }
            }

            if (m_queued_keyword_refresh)
            {
                RefreshKeywords();

                m_queued_keyword_refresh = false;
            }
        }
    }
}
